#ifndef __DLL_HH
#define __DLL_HH

//
//  Project   : misc
//  File      : DLL.hh
//  Author    : Ronald Kriemann
//  Purpose   : Template class for a double concatenated list
//

#include <assert.h>

namespace LWL
{

////////////////////////////////////////////////////
// 
// Template class list
//
template <class T> 
class DLL
{
public:
    ////////////////////////////////////////////////////
    //
    // class for a single item in the list
    //

    class TItem
    {
    public:
        T        _data;
        TItem  * _next, * _prev;
        
    public:
        TItem ( const T  & e, 
                TItem    * n = NULL,
                TItem    * p = NULL )
                : _data( e ), _next( n ), _prev( p )
        {}
    };

    //////////////////////////////////////////////////////////
    // 
    // class for an iterator for the previous list
    //

    class TIterator
    {
        friend class DLL<T>;
        
    protected:
        // currently selected item
        TItem  * _item;
        
    public:
        //
        // constructor and destructor
        //

        TIterator ()                         : _item(NULL) {}
        TIterator ( TItem * item )           : _item(item) {}
        TIterator ( const TIterator & iter ) : _item(iter._item) {}
        
        // get status (end-of-list)
        bool eol () const { return (_item == NULL); }
        
        // increment
        TIterator & next () { return ++*this; }
        TIterator & prev () { return --*this; }
        
        // access actual element
        T &       data ()       { assert(_item != NULL); return _item->_data; }
        const T & data () const { assert(_item != NULL); return _item->_data; }

        //
        // usual operators
        //

        // copy
        TIterator & operator = ( const TIterator & i ) { _item = i._item; return *this; }

        // access items
        T &         operator () ()       { assert(_item != NULL); return _item->_data; }
        const T &   operator () () const { assert(_item != NULL); return _item->_data; }

        // iterate (prefix/postfix)
        TIterator & operator ++ ()    { if (_item) _item = _item->_next; return *this; }
        TIterator   operator ++ (int) { TIterator tmp(*this); ++*this; return tmp; }

        TIterator & operator -- ()    { if (_item) _item = _item->_prev; return *this; }
        TIterator   operator -- (int) { TIterator tmp(*this); --*this; return tmp; }

        // dereference
        T &         operator *  ()    { assert(_item != NULL); return   _item->_data; }
        T *         operator -> ()    { assert(_item != NULL); return & _item->_data; }

        // compare
        bool        operator == ( const TIterator & i ) const { return (_item == i._item); }
        bool        operator != ( const TIterator & i ) const { return (_item != i._item); }
    };

protected:
    // first and last element of list
    TItem        * _first, * _last;
    
    // size of the list aka number of stored items
    unsigned int   _size;

public:
    /////////////////////////////////////////////////
    //
    // constructors and destructors
    //

    DLL ()                   : _first(NULL), _last(NULL), _size(0) {}
    DLL ( const DLL<T> & l ) : _first(NULL), _last(NULL), _size(0) { copy( l ); }

    virtual ~DLL () { if (_size > 0) remove_all(); }

    /////////////////////////////////////////////////
    //
    // access member variables
    //

    unsigned int size () const { return _size; }

    TIterator       first ()       { return _first; }
    const TIterator first () const { return _first; }

    TIterator       last  ()       { return _last; }
    const TIterator last  () const { return _last; }

    // same as above but access real data
    T &       first_item ()       { assert(_first != NULL); return _first->_data; }
    const T & first_item () const { assert(_first != NULL); return _first->_data; }
    
    T &       last_item ()       { assert(_last != NULL); return _last->_data; }
    const T & last_item () const { assert(_last != NULL); return _last->_data; }

    /////////////////////////////////////////////////
    //
    // methods for manipulating list
    //

    // add a new object to the begin/end of list
    DLL<T> & prepend ( const T & elem );
    DLL<T> & append  ( const T & elem );

    // remove (next) element from list
    void remove ( const T & elem );
    void remove ( TIterator & i ) { assert(i._item != NULL); remove(i._item->_data); }
    void remove_first ();
    void remove_last ();

    // return first item and remove it from list
    T behead ();
    
    // remove all elements from list
    void remove_all ();

    // copy given list into local
    void copy ( const DLL< T > & list );
    
    /////////////////////////////////////////////////
    //
    // search/sort related methods
    //

    // look if element is in list
    bool is_in ( const T & elem );

    /////////////////////////////////////////////////
    //
    // misc.
    //
    
    // return size in bytes used by this object
    unsigned long byte_size () const;
};

#include "DLL.cc"

}// namespace

#endif   // __DLL_HH
